/// Format reports using this lightweight Dart Library.
/// Supports Dart and Flutter. Flutter applications may
/// want to use the StreamBuffer option.
/// 

library flexprint;


enum Overflow {
  truncate,
  wrap,
}

enum Alignment {
  start,
  center,
  end,
}

class FlexPrint {
  final int maxWidth;
  final int minPadding;
  final bool useStringbuffer;
  StringBuffer _stringBuffer;
  Overflow overflow;

  List<Map<String, dynamic>> chunks = [];

  /// **FlexPrint** - uses the flex layout system to print
  /// columns in a row.
  ///
  /// A row is filled by repeated calls to *write()* method for each column
  /// if flex is not specified flex = 1 is assumed.
  ///
  /// A new line **\n** character causes the current row to be printed.
  /// If there are multiple new line characters, muiltiple rows are printed
  /// If there is no newline character - the row is not printed. The row
  /// can be output by calling write with a single newline.
  ///
  /// If **useStringbuffer** is set to *true*, FlexPrint will print to
  /// a stringBuffer instead of console.
  ///
  ///*FlexPrint.stringBuffer* will return the stringbuffer being used
  /// FlexPrint. It is the user's responsibility to clear the stringbuffer
  /// if FlexPrint object is being reused.

  FlexPrint({
    this.maxWidth = 80,
    this.minPadding = 1,
    this.overflow = Overflow.truncate,
    this.useStringbuffer = false,
  }) : _stringBuffer = useStringbuffer ? StringBuffer() : null;

  StringBuffer get stringBuffer => _stringBuffer;

  /// write column chunk, flex (defaults to 1). *Flex* determines
  /// the number of parts of the row this column shall occupy. Think
  /// of *flex* like a fraction. If the row has 10 columns and this
  /// column has a flex = 3, then the column shall occupy 3/10ths or
  /// 30% of the width of the row.
  ///
  /// The chunk can not be empty. If the chunk is empty or null, FlexPrint
  /// will simply move next column in it's place, causing columns to be out
  /// of alignment with other rows.
  ///
  /// A *newline* will **end** current row and print current row. Any text after
  /// the newline will be in the next row.
  ///
  /// The column contents may be **aligned** with start, center or end
  /// of column by specifying: *Alignment.start,Alignment.center or
  /// Alignkment.end*.
  ///

  void write(String chunk,
      {int flex = 1, Alignment alignment = Alignment.start}) {
    if (chunk.contains('\n')) {
      final List<String> chunkParts = chunk.split('\n');
      chunks.add({
        'chunk': chunkParts[0],
        'flex': flex == 0 ? 1 : flex,
        'alignment': alignment
      });
      _print();
      if (chunkParts.length > 1) {
        chunkParts.removeAt(0);
        write(chunkParts.join('\n'),
            flex: flex == 0 ? 1 : flex, alignment: alignment);
      }
    } else {
      chunks.add({
        'chunk': chunk,
        'flex': flex == 0 ? 1 : flex,
        'alignment': alignment
      });
    }
  }

  /// prints a horizontal line to the whole width.
  /// The line is drawn using default char  **-**, you may choose any other char.
  /// The default is a single line char. By specifying *doubleLine = true*,
  /// a double line is drawn with **=**.

  void horizontalLine({String char = '-', bool doubleLine = false}) {
    String s = '';
    for (int i = 0; i < maxWidth + 1; i++) {
      s += doubleLine ? '=' : char.substring(0, 1);
    }
    write('$s\n');
  }

  String _toString() {
    int totalFlex = 0;
    chunks.removeWhere((chunk) => chunk['chunk'].isEmpty);
    chunks.forEach((chunk) {
      totalFlex += chunk['flex'];
    });
    int flexfactor = maxWidth ~/ totalFlex;
    List<StringBuffer> sbs = [];
    chunks.forEach((chunk) {
      final fb = _foldChunk(
          chunk['chunk'], flexfactor * chunk['flex'], chunk['alignment']);
      for (int i = 0; i < fb.length; i++) {
        if (sbs.isEmpty || sbs.length < i) {
          sbs.add(StringBuffer());
        }
        sbs[i].write(fb[i]);
      }
    });
    String str = '';
    sbs.forEach((sb) {
      str += (sb.toString() + '\n');
    });
    sbs.forEach((sb) {
      sb.clear();
    });
    return str.substring(0, str.lastIndexOf('\n'));
  }

  String _padding(int length) {
    String s = '';
    for (int i = 0; i < length; i++) {
      s += ' ';
    }
    return s;
  }

  List<String> _foldChunk(String chunk, int width, Alignment alignment) {
    final List<String> buffer = [];

    String first = chunk.length > width ? chunk.substring(0, width) : chunk;
    buffer.add(_align(first, width, alignment));
    if (overflow == Overflow.truncate) {
      return buffer;
    }
    if (chunk.length > width) {
      _foldChunk(chunk.substring(width), width, Alignment.start);
    }
    return buffer;
  }

  String _align(String text, int width, Alignment alignment) {
    //print('$chunk  ${width.toString()}');
    final chunk = text + _padding(minPadding);
    switch (alignment) {
      case Alignment.center:
        var pad = ((width - chunk.length) / 2).round();
        int rightPad = width - (chunk.length + pad);
        return _padding(pad) + chunk + _padding(rightPad);
      case Alignment.end:
        return _padding(width - chunk.length) + chunk;
      default:
        return chunk + _padding(width - chunk.length);
    }
  }

  void _print() {
    if (useStringbuffer) {
      _stringBuffer.write(toString());
    } else {
      print(_toString());
    }
    chunks = [];
  }
}
